#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#ifndef GL_GLEXT_PROTOTYPES
#define GL_GLEXT_PROTOTYPES
#endif

#include "gl_util.h"
#include "util.h"
#include "globals.h"

using namespace globals;

void init_gl() {
   glShadeModel(GL_SMOOTH);
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glClearDepth(1);
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LESS);

   glEnable(GL_LIGHT0);
   glEnable(GL_TEXTURE_2D);
}

void glut_resize(int w, int h) {
   glViewport(0, 0, w, h); 
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   gluPerspective(cam::field_of_view, w / (float) h, cam::near_plane_dist, cam::far_plane_dist);
   glMatrixMode(GL_MODELVIEW);
}

void center_cursor() {
   glutWarpPointer(main_win::width >> 1, main_win::height >> 1);
}

// Taken from ZJW's 473 Lab 5 sample code
int printOglError (const char *file, int line) {
   /* Returns 1 if an OpenGL error occurred, 0 otherwise. */
   GLenum glErr;
   int    retCode = 0;

   glErr = glGetError ();
   while (glErr != GL_NO_ERROR)
   {
      printf ("glError in file %s @ line %d: %s\n", file, line, gluErrorString (glErr));
      retCode = 1;
      glErr = glGetError ();
   }
   return retCode;
}

// Taken from ZJW's 473 Lab 5 sample code
void printShaderInfoLog (GLuint shader)
{
   GLint     infologLength = 0;
   GLint     charsWritten  = 0;
   GLchar *infoLog;

   printOpenGLError ();  // Check for OpenGL errors
   glGetShaderiv (shader, GL_INFO_LOG_LENGTH, &infologLength);
   printOpenGLError ();  // Check for OpenGL errors

   if (infologLength > 0)
   {
      infoLog = (GLchar *)malloc (infologLength);
      if (infoLog == NULL)
      {
         puts ("ERROR: Could not allocate InfoLog buffer");
         exit (1);
      }
      glGetShaderInfoLog (shader, infologLength, &charsWritten, infoLog);
      if (strcmp(infoLog, "") != 0)
         printf ("Shader InfoLog:\n%s\n\n", infoLog);
      free (infoLog);
   }
   printOpenGLError();  // Check for OpenGL errors
}

// Taken from ZJW's 473 Lab 5 sample code
void printProgramInfoLog (GLuint program)
{
   GLint     infologLength = 0;
   GLint     charsWritten  = 0;
   GLchar *infoLog;

   printOpenGLError ();  // Check for OpenGL errors
   glGetProgramiv (program, GL_INFO_LOG_LENGTH, &infologLength);
   printOpenGLError ();  // Check for OpenGL errors

   if (infologLength > 0)
   {
      infoLog = (GLchar *)malloc (infologLength);
      if (infoLog == NULL)
      {
         puts ("ERROR: Could not allocate InfoLog buffer");
         exit (1);
      }
      glGetProgramInfoLog (program, infologLength, &charsWritten, infoLog);
      if (strcmp(infoLog, "") != 0)
         printf ("Program InfoLog:\n%s\n\n", infoLog);
      free (infoLog);
   }
   printOpenGLError ();  // Check for OpenGL errors
}

// Taken from ZJW's 473 Lab 5 sample code
GLint getUniLoc(GLuint program, const GLchar *name) {
   GLint loc;

   loc = glGetUniformLocation(program, name);

   if (loc == -1) {
      fprintf(stderr, "No such uniform named %s in program %d\n", name, program);
   }

   printOpenGLError();
   return loc;
}

GLint getAttribLoc(GLuint program, const GLchar *name) {
   GLint loc;

   loc = glGetAttribLocation(program, name);

   if (loc == -1) {
      fprintf(stderr, "No such attribute named %s in program %d\n", name, program);
   }

   printOpenGLError();
   return loc;
}

void draw_base_grid() {
   glDisable(GL_LIGHTING);
   glBegin(GL_LINES); {
      glColor3f(0, 1, 0);
      for (float i = -12.0; i <= 12.0f; i += 1) {

         glVertex3f(i, 0, -12.0f);
         glVertex3f(i, 0, 12.0f);

         glVertex3f(-12.0f, 0, i);
         glVertex3f(12.0f, 0, i);
      }
   }
   glEnd();
   glEnable(GL_LIGHTING);
}

static int compile_shader(const char* path, unsigned int type) {
   int ret = glCreateShader(type);
   char* shader_text = slurp_file(path);
   const GLchar* pass_text = (const GLchar*) shader_text;
   int status; 

   glShaderSource(ret, 1, &pass_text, NULL);

   printOglError(__FILE__, __LINE__);
   glGetShaderiv(ret, GL_COMPILE_STATUS, &status);
   if (status == GL_FALSE)
      printShaderInfoLog(ret);

   free(shader_text);
   return ret;
}

// Compiles a vertex shader from the file at "path"
int compile_vertex_shader(const char* path) {
   compile_shader(path, GL_VERTEX_SHADER);
}

// Compiles a fragment shader from the file at "path"
int compile_fragment_shader(const char* path) {
   compile_shader(path, GL_FRAGMENT_SHADER);
}

// Links a vertex & fragment shader together to make a valid program
int compile_program(int vshader, int fshader) {
   int ret = glCreateProgram();
   int status;

   glAttachShader(ret, vshader);
   glAttachShader(ret, fshader);
   glLinkProgram(ret);
   printOpenGLError();

   glGetProgramiv(ret , GL_LINK_STATUS, &status);
   if (status == GL_FALSE)
      printProgramInfoLog(ret);
   return ret;
}

void activate_program(int prog) {
   assert(prog >= 0);
   glUseProgram(prog);
}

void deactivate_shaders() {
   glUseProgram(0);
}
